var {
  THREE,
  Edge,
  Node,
  OrbitControls,
  d3,
  ee,
  DEF_NODE_COLOR,
  DEF_NODE_HIGHLIGHT,
  DEF_ADJACENT_HIGHLIGHT,
  DEF_NODE_HTEXTURE,
  DEF_NODE_TEXTURE,
  DEF_NODE_NO_HTEX,
  DEF_NODE_NO_TEX,
  DEF_NODE_NO_HCOLOR,
  DEF_NODE_NO_COLOR,
  DEF_NODE_NO_HIGHLIGHT,
  DEF_NODE_SHAPE,
  DEF_NODE_HSHAPE,
  DEF_NODE_SIZE,
  DEF_HIGHLIGHT_SIZE,
  DEF_LAYOUT,
  DEF_LINES,
  DEF_MAP,
  DEF_TEXT_SCALE,
  DEF_TEXT_SIZE,
  DEF_TEXT_X_OFFSET,
  DEF_TEXT_Y_OFFSET,
  DEF_TEXT_BACK_COLOR,
  DEF_TEXT_OPACITY,
  TEXT_WIREFRAME
} = require('./imports').default;

module.exports = function (self) {
  var checked = true;

  /**
   *  on Mouse Move
   */
  self.onMouseMove = function (selection, mouseX, mouseY, button, ctrl) {

    // mouse moves while no button is clicked
    if (!self.mouseDown) {
      self.onHover(selection);
      self.mouseStart = new THREE.Vector3(mouseX, mouseY, 0);

    // left button clicked
    } else if (button == 0) {

      // left-clicked empty space (i.e., not clicking a node)
      if (!self.dragging && self.selection.indexOf(selection) == -1 && !ctrl) {
        // clear node selection
        self.clearSelection();
      }

      // add nodes enclosed by selection box into node selection
      self.checkSelection(mouseX, mouseY);
      
      // control key not pressed 
      if (!ctrl) {

        // at least one node selected
        if (self.selection.length > 0) {

          // at this point: left button clicked, one or more nodes selected and dragged
          // reactivate (in D3's terminology: reheat) the force layout so that as user
          // drags nodes, other nodes will move along
          self.force.alpha(1);


          for (var i = 0; i < self.selection.length; i++) {

            self.selection[i].renderData.isSelected = false;

            if (!DEF_NODE_NO_HIGHLIGHT) {
              self.selection[i].renderData.draw_object.children[0].visible = false;
            } else {
              self.selection[i].renderData.draw_object.material.color.set(new THREE.Color(self.selection[i].renderData.color));
            }

            self.selection[i].renderData.textHolder.children[0].element.hideme = true;
          }
        }
      }

      if (self.selection.length > 0) {
        self.updateSelection(mouseX, mouseY);
      }

      checked = true;

      if (!self.dragging) {
        
        if (self.showBox) {
          self.selectBox.visible = true;
          self.showBox = false;
        }

        self.selectBox.position.x = mouseX;
        self.selectBox.position.y = mouseY;
        var diffx = self.mouseStart.x - mouseX;
        var diffy = self.mouseStart.y - mouseY;
        self.selectBox.scale.set(diffx, diffy, 1);
      } else {
        self.selectBox.visible = false;
      }
    }
  }

  /**
   * clear selection
   */
  self.clearSelection = function () {
    for (var i = 0; i < self.selection.length; i++) {
      self.selection[i].fx = null;
      self.selection[i].fy = null;
      self.selection[i].renderData.isSelected = false;
      self.selection[i].renderData.isDragging = false;
      if (!DEF_NODE_NO_HIGHLIGHT) {
        self.selection[i].renderData.draw_object.children[0].visible = false;
      } else {
        self.selection[i].renderData.draw_object.material.color.set(new THREE.Color(self.selection[i].renderData.color));
      }
      self.selection[i].renderData.textHolder.children[0].element.hideme = true;
    }
    self.selection = [];
  }

  /**
   * on Hover
   */
  self.onHover = function (node) {
    if (self.lastHover && self.selection.indexOf(self.lastHover) == -1) {
      if (!DEF_NODE_NO_HIGHLIGHT) {
        self.lastHover.renderData.draw_object.children[0].visible = false;
      } else {
        self.lastHover.renderData.draw_object.material.color.set(new THREE.Color(self.lastHover.renderData.color));
      }
      self.lastHover.renderData.textHolder.children[0].element.hideme = true;
      const froms = [];
      const tos = [];
      for (var i = 0; i < self.selection.length; i++) {
        for (var j = 0; j < self.selection[i].links.length; j++) {
          froms.push(self.selection[i].links[j].fromId);
          tos.push(self.selection[i].links[j].toId);
        }
      }
      self.graph.forEachNode((n) => {
        if (self.selection.indexOf(n) != -1
          || n == node) {
          n.renderData.draw_object.material.opacity = 1;
          n.renderData.linecolor.r = 0.5;
          n.renderData.linecolor.g = 0.5;
          n.renderData.linecolor.b = 0.5;
          for (var i = 0; i < n.linkObjs.length; i++) {
            n.linkObjs[i].linecolor = n.renderData.linecolor;
          }
        } else if (froms.indexOf(n.id) != -1
          || tos.indexOf(n.id) != -1) {
          n.renderData.draw_object.material.opacity = 1;
          n.renderData.linecolor.r = 0.9;
          n.renderData.linecolor.g = 0.9;
          n.renderData.linecolor.b = 0.9;
          if (n.renderData.draw_object.children[0]) {
            n.renderData.draw_object.children[0].material.color.set(DEF_ADJACENT_HIGHLIGHT);
            n.renderData.draw_object.children[0].visible = true;
          }
        } else {
          n.renderData.draw_object.material.opacity = 0.5;
          n.renderData.linecolor.r = 0.9;
          n.renderData.linecolor.g = 0.9;
          n.renderData.linecolor.b = 0.9;
          n.renderData.draw_object.children[0].material.color.setHex(n.renderData.hcolor);
          n.renderData.draw_object.children[0].visible = false;
        }
      });
    }
    self.lastHover = node;
    if (node) {
      if (!DEF_NODE_NO_HIGHLIGHT) {
        node.renderData.draw_object.children[0].visible = true;
      } else {
        node.renderData.draw_object.material.color.setHex(node.renderData.hcolor);
      }
      node.renderData.textHolder.children[0].element.hideme = false;
      const froms = [];
      const tos = [];
      for (var i = 0; i < node.links.length; i++) {
        froms.push(node.links[i].fromId);
        tos.push(node.links[i].toId);
      }
      for (var i = 0; i < self.selection.length; i++) {
        for (var j = 0; j < self.selection[i].links.length; j++) {
          froms.push(self.selection[i].links[j].fromId);
          tos.push(self.selection[i].links[j].toId);
        }
      }
      self.graph.forEachNode((n) => {
        if (self.selection.indexOf(n) != -1
          || n == node) {
          n.renderData.draw_object.material.opacity = 1;
          n.renderData.linecolor.r = 0.5;
          n.renderData.linecolor.g = 0.5;
          n.renderData.linecolor.b = 0.5;
          for (var i = 0; i < n.linkObjs.length; i++) {
            n.linkObjs[i].linecolor = n.renderData.linecolor;
          }
        } else if (froms.indexOf(n.id) != -1
          || tos.indexOf(n.id) != -1) {
          n.renderData.draw_object.material.opacity = 1;
          n.renderData.linecolor.r = 0.9;
          n.renderData.linecolor.g = 0.9;
          n.renderData.linecolor.b = 0.9;
          if (n.renderData.draw_object.children[0]) {
            n.renderData.draw_object.children[0].material.color.set(DEF_ADJACENT_HIGHLIGHT);
            n.renderData.draw_object.children[0].visible = true;
          }
        } else {
          n.renderData.draw_object.material.opacity = 0.5;
          n.renderData.linecolor.r = 0.9;
          n.renderData.linecolor.g = 0.9;
          n.renderData.linecolor.b = 0.9;
          n.renderData.draw_object.children[0].material.color.setHex(n.renderData.hcolor);
          n.renderData.draw_object.children[0].visible = false;
        }
      });
    } else if (self.selection.length == 0) {
      self.graph.forEachNode((n) => {
        n.renderData.draw_object.material.opacity = 1;
        n.renderData.linecolor.r = 0.5;
        n.renderData.linecolor.g = 0.5;
        n.renderData.linecolor.b = 0.5;
        n.renderData.draw_object.children[0].material.color.setHex(n.renderData.hcolor);
        n.renderData.draw_object.children[0].visible = false;
      });
    }
  }

  /**
   *  on Mouse Down
   */
  self.onMouseDown = function (selection, mouseX, mouseY, button, ctrl) {
    self.mouseDown = true;
    if (button == 0) {      
      self.mouseStart = new THREE.Vector3(mouseX, mouseY, 0);
      if (!self.dragging) {
        self.showBox = true;
      }
      if (self.selection.indexOf(selection) == -1 && !ctrl) {
        for (var i = 0; i < self.selection.length; i++) {
          self.selection[i].renderData.isSelected = false;
          self.selection[i].renderData.isDragging = false;
          if (!DEF_NODE_NO_HIGHLIGHT) {
            self.selection[i].renderData.draw_object.children[0].visible = false;
          } else {
            self.selection[i].renderData.draw_object.material.color.set(new THREE.Color(self.selection[i].renderData.color));
          }
          self.selection[i].renderData.textHolder.children[0].element.hideme = true;
        }
        self.selection = [];
      }
      if (selection) {
        self.dragging = selection;
        if (!selection.renderData.isSelected) {
          if (self.selection.indexOf(selection) == -1) {
            self.selection.push(selection);
          }
          selection.renderData.isSelected = true;
          self.updateSelection(mouseX, mouseY);
        } else if (ctrl) {
          self.selection.splice(self.selection.indexOf(selection), 1);
          selection.renderData.isSelected = false;
          if (!DEF_NODE_NO_HIGHLIGHT) {
            selection.renderData.draw_object.children[0].visible = false;
          } else {
            selection.renderData.draw_object.material.color.set(new THREE.Color(self.selection[i].renderData.color));
          }
          selection.renderData.textHolder.children[0].element.hideme = true;
          self.dragging = null;
        }
      }
    }
  }

  /**
   *  on Mouse Up
   */
  self.onMouseUp = function (selection, mouseX, mouseY, button) {
    self.mouseDown = false;
    if (button == 0) {
      self.showBox = false;
      self.dragging = null;
      self.selectBox.visible = false;
      if(selection)
        selection.renderData.isDragging = false;
      
      self.onSelectNodes();
      
      self.ee.emit('select-nodes', self.selection);
    }
  }

  /**
   *  find radius of circle given the volume
   */
  self.volToRadius = function (volume) {
    return Math.pow((volume / Math.PI) * (3.0 / 4.0), 1.0 / 3.0)
  }

  /**
   *  on Right Click
   */
  self.onRightClick = function (selection) {
    // self.inGraph.forEachNode((n) => {
    //   n.data.size = 4;
    // });
    // self.updateGraph(self.inGraph);
    //self.setLabelLength(8);
    //self.removeNode(selection);
    //self.setLabelFontSize(10);
    // if (!self.pinned) {
    //   self.pinSelectedNodes();
    //   self.pinned = true;
    // } else {
    //   self.unpinSelectedNodes();
    //   self.pinned = false;
    // }
    // var createNGraph = require('ngraph.graph');
    // var newGraph = createNGraph();
    // self.graph.forEachNode(function (node) {
    //   newGraph.addNode(node.id, node.data);
    // });
    // self.graph.forEachLink(function (link) {
    //   newGraph.addLink(link.fromId, link.toId);
    // });
    // newGraph.addNode("952", {label: "Test", x: 0, y: 0, size:50});
    // newGraph.addNode("953", {label: "Test", x: 0, y: 0, size:50});
    // newGraph.addNode("955", {label: "Test", x: 0, y: 0, size:50});
    // newGraph.addNode("956", {label: "Test", x: 0, y: 0, size:50});
    // newGraph.addLink("952", "953");
    // newGraph.addLink("955", "956");
    // newGraph.addLink("955", "953");
    // self.updateGraph(newGraph);
    // if (self.paused) {
    //   self.resumeLayout();
    // } else {
    //   self.pauseLayout();
    // }
    // self.mapNodeAttributes(["scalex", "", "labelSize"])
    // self.mapAllNodes();
    // self.mapNodeAttributes(["degree", "log", "scalexy"], self.selection);
  }

  /**
   *  Update positions of all objects in self.selection
   *  based on diff between mouse position and self.dragging position
   */
  self.updateSelection = function (mouseX, mouseY) {
    if (self.dragging) {
      var diffx = mouseX - self.dragging.x;
      var diffy = mouseY - self.dragging.y;
    }
    for (var i = 0; i < self.selection.length; i++) {
      if (self.dragging) {
        self.selection[i].x += diffx;
        self.selection[i].y += diffy;
      }
      self.selection[i].fx = self.selection[i].x;
      self.selection[i].fy = self.selection[i].y;
      self.selection[i].renderData.isDragging = true;
      if (!DEF_NODE_NO_HIGHLIGHT) {
        self.selection[i].renderData.draw_object.children[0].visible = true;
      } else {
        self.selection[i].renderData.draw_object.material.color.set(new THREE.Color(self.selection[i].renderData.hcolor));
      }
      self.selection[i].renderData.textHolder.children[0].element.hideme = false;
    }
  }

  /**
   *  Find any objects within the current box selection and add it to self.selection
   */
  self.checkSelection = function (mouseX, mouseY) {
    if (!self.dragging) {
      self.mouseEnd = new THREE.Vector3(mouseX, mouseY, 0);
      if (self.mouseStart.x < self.mouseEnd.x) {
        var left = self.mouseStart;
        var right = self.mouseEnd;
      } else {
        var left = self.mouseEnd;
        var right = self.mouseStart;
      }

      self.graph.forEachNode(function (node) {
        let npos
        if (self.options.layout == "ngraph") {
          npos = self.force.getNodePosition(node.id);
        } else if (self.options.layout == "d3") {
          npos = node;
        }
        if (self.insideBox(left, right, npos.x, npos.y)) {
          self.selection.push(node);
        }
      });
    }
  }

  /**
   *  returns if pos is in box with top left l and bottom right r
   */
  self.insideBox = function (l, r, posx, posy) {
    return posx < r.x && posx > l.x && ((posy > r.y && posy < l.y) || (posy < r.y && posy > l.y))
  }
}